/*
 * pama-applet.c: Main interface
 * Part of PulseAudio Mixer Applet
 * Copyright © Vassili Geronimos 2009 <v.geronimos@gmail.com>
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif
#include <glib.h>
#include <glib/gi18n.h>
 
#include <string.h>
#include "pama-applet.h"

#define VOLUME_SCROLL_STEP (PA_VOLUME_NORM/20)

static gboolean pama_applet_delayed_init              (gpointer data);
static void     pama_applet_dispose                   (GObject *gobject);
static void     pama_applet_change_orient             (PanelApplet *applet, PanelAppletOrient orient);
static void     pama_applet_pulse_context_disconnected(PamaPulseContext *context, gpointer data);
static void     pama_applet_pulse_context_connected   (PamaPulseContext *context, gpointer data);
static void     pama_applet_pulse_context_notify      (PamaPulseContext *context, GParamSpec *pspec, gpointer data);
static void     pama_applet_update_default_devices    (PamaApplet *applet);
static void     pama_applet_io_devs_changed           (gpointer, guint, gpointer data);
static void     pama_applet_io_devs_notify            (gpointer, GParamSpec *, gpointer data);
static gboolean pama_applet_create_context            (gpointer data);
static void     pama_applet_update_icons              (PamaApplet *applet);

static gboolean pama_applet_icon_scroll  (GtkWidget *event_box, GdkEventScroll *event, gpointer data);
static gboolean pama_applet_icon_click   (GtkWidget *event_box, GdkEventButton *event, gpointer data);

static void pama_applet_show_about_dialog(BonoboUIComponent *uic, gpointer data, const char *cname);
static const BonoboUIVerb pama_applet_menu_verbs[] = {
	BONOBO_UI_VERB("About", pama_applet_show_about_dialog),
	BONOBO_UI_VERB_END
};

G_DEFINE_TYPE(PamaApplet, pama_applet, PANEL_TYPE_APPLET);

static void pama_applet_class_init(PamaAppletClass *klass)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
	PanelAppletClass *applet_class = PANEL_APPLET_CLASS(klass);

	gobject_class->dispose  = pama_applet_dispose;
	applet_class->change_orient = pama_applet_change_orient;
}

static void pama_applet_init(PamaApplet *self)
{
	self->orientation = panel_applet_get_orient(PANEL_APPLET(self));

	if (pama_applet_is_horizontal(self->orientation))
		self->box = GTK_BOX(gtk_hbox_new(FALSE, 2));
	else
		self->box = GTK_BOX(gtk_vbox_new(FALSE, 2));

	gtk_container_add(GTK_CONTAINER(self), GTK_WIDGET(self->box));

	self->sink_event_box   = gtk_event_box_new();
	self->source_event_box = gtk_event_box_new();
	gtk_event_box_set_visible_window(GTK_EVENT_BOX(self->sink_event_box),   FALSE);
	gtk_event_box_set_visible_window(GTK_EVENT_BOX(self->source_event_box), FALSE);
	gtk_widget_set_tooltip_text(GTK_WIDGET(self->sink_event_box),   _("Not connected to PulseAudio server"));
	gtk_widget_set_tooltip_text(GTK_WIDGET(self->source_event_box), _("Not connected to PulseAudio server"));
	gtk_widget_set_sensitive(GTK_WIDGET(self->sink_event_box),   FALSE);
	gtk_widget_set_sensitive(GTK_WIDGET(self->source_event_box), FALSE);
	gtk_box_pack_start(self->box, self->sink_event_box,   FALSE, TRUE, 0);
	gtk_box_pack_start(self->box, self->source_event_box, FALSE, TRUE, 0);

	self->sink_icon   = gtk_image_new_from_icon_name("audio-volume-muted",           GTK_ICON_SIZE_MENU);
	self->source_icon = gtk_image_new_from_icon_name("audio-input-microphone-muted", GTK_ICON_SIZE_MENU);
	gtk_container_add(GTK_CONTAINER(self->sink_event_box),   self->sink_icon);
	gtk_container_add(GTK_CONTAINER(self->source_event_box), self->source_icon);

	g_idle_add(pama_applet_delayed_init, self);
	g_object_connect(self->sink_event_box,
	                 "signal::scroll-event",       pama_applet_icon_scroll,    self,
	                 "signal::button-press-event", pama_applet_icon_click,     self,
	                 NULL);
	g_object_connect(self->source_event_box,
	                 "signal::scroll-event",       pama_applet_icon_scroll,    self,
	                 "signal::button-press-event", pama_applet_icon_click,     self,
	                 NULL);

	panel_applet_set_flags(PANEL_APPLET(self), PANEL_APPLET_EXPAND_MINOR);
	panel_applet_set_background_widget(PANEL_APPLET(self), GTK_WIDGET(self));
	gtk_widget_show_all(GTK_WIDGET(self));
}

static gboolean pama_applet_delayed_init(gpointer data)
{
	panel_applet_setup_menu_from_file(PANEL_APPLET(data),
	                                  DATADIR,
	                                  "PulseaudioMixerApplet.xml",
	                                  NULL,
	                                  pama_applet_menu_verbs,
	                                  data);

	return FALSE;
}

void pama_applet_set_api(PamaApplet *self, pa_mainloop_api *api)
{
	g_warn_if_fail(api);

	self->api = api;
	g_idle_add(pama_applet_create_context, self);
}

static gboolean pama_applet_create_context(gpointer data)
{
	PamaApplet *self = data;

	self->context = pama_pulse_context_new(self->api);
	g_object_connect(self->context, 
	                 "signal::disconnected",   pama_applet_pulse_context_disconnected, self,
	                 "signal::connected",      pama_applet_pulse_context_connected,    self,
					 "signal::notify",         pama_applet_pulse_context_notify,       self,
					 "signal::sink-added",     pama_applet_io_devs_changed,            self,
					 "signal::sink-removed",   pama_applet_io_devs_changed,            self,
					 "signal::source-added",   pama_applet_io_devs_changed,            self,
					 "signal::source-removed", pama_applet_io_devs_changed,            self,
	                 NULL);

	return FALSE;
}

static void pama_applet_pulse_context_connected(PamaPulseContext *context, gpointer data)
{
	PamaApplet *self = data;

	gtk_widget_set_sensitive(GTK_WIDGET(self->sink_event_box),   TRUE);
	gtk_widget_set_sensitive(GTK_WIDGET(self->source_event_box), TRUE);
	gtk_widget_set_tooltip_text(GTK_WIDGET(self->sink_event_box),   _("No information received yet"));
	gtk_widget_set_tooltip_text(GTK_WIDGET(self->source_event_box), _("No information received yet"));
}

static void pama_applet_pulse_context_disconnected(PamaPulseContext *context, gpointer data)
{
	PamaApplet *self = data;

	g_object_unref(context);

	self->default_sink = NULL;
	self->default_source = NULL;
	self->context = NULL;
	
	if (self->sink_popup)
		gtk_widget_destroy(GTK_WIDGET(self->sink_popup));
	if (self->source_popup)
		gtk_widget_destroy(GTK_WIDGET(self->source_popup));

	gtk_image_set_from_icon_name(GTK_IMAGE(self->sink_icon),   "audio-volume-muted",           GTK_ICON_SIZE_MENU);
	gtk_image_set_from_icon_name(GTK_IMAGE(self->source_icon), "audio-input-microphone-muted", GTK_ICON_SIZE_MENU);
	gtk_widget_set_tooltip_text(GTK_WIDGET(self->sink_event_box),   _("Not connected to PulseAudio server"));
	gtk_widget_set_tooltip_text(GTK_WIDGET(self->source_event_box), _("Not connected to PulseAudio server"));
	gtk_widget_set_sensitive(GTK_WIDGET(self->sink_event_box),   FALSE);
	gtk_widget_set_sensitive(GTK_WIDGET(self->source_event_box), FALSE);

	g_timeout_add_seconds(10, pama_applet_create_context, self);
}

static void pama_applet_pulse_context_notify(PamaPulseContext *context, GParamSpec *pspec, gpointer data)
{
	PamaApplet *self = data;

	pama_applet_update_default_devices(self);
}

static void pama_applet_io_devs_changed(gpointer blah, guint index, gpointer data)
{
	/* Handle (sink|source)-(added|removed) signals */
	PamaApplet *self = data;

	pama_applet_update_default_devices(self);
	pama_applet_update_icons(self);
}

static void pama_applet_io_devs_notify(gpointer blah, GParamSpec *pspec, gpointer data)
{
	PamaApplet *self = data;

	pama_applet_update_icons(self);
}

static void pama_applet_update_default_devices(PamaApplet *self)
{
	PamaPulseSink    *new_default_sink   = pama_pulse_context_get_default_sink  (self->context);
	PamaPulseSource  *new_default_source = pama_pulse_context_get_default_source(self->context);

	if(new_default_sink != self->default_sink)
	{
		if (self->default_sink)
			g_signal_handlers_disconnect_by_func(self->default_sink, pama_applet_io_devs_notify, self);

		self->default_sink = new_default_sink;

		if (self->default_sink)
			g_signal_connect(self->default_sink, "notify::volume", G_CALLBACK(pama_applet_io_devs_notify), self);
	}

	if(new_default_source != self->default_source)
	{
		if (self->default_source)
			g_signal_handlers_disconnect_by_func(self->default_source, pama_applet_io_devs_notify, self);

		self->default_source = new_default_source;

		if (self->default_source)
			g_signal_connect(self->default_source, "notify::volume", G_CALLBACK(pama_applet_io_devs_notify), self);
	}

	pama_applet_update_icons(self);
}

static void pama_applet_change_orient(PanelApplet *applet, PanelAppletOrient orientation)
{
	PamaApplet *self = PAMA_APPLET(applet);
	GtkBox *box;

	if (pama_applet_is_horizontal(self->orientation) == pama_applet_is_vertical(orientation))
	{
		if (pama_applet_is_vertical(orientation))
			box = GTK_BOX(gtk_vbox_new(FALSE, 2));
		else
			box = GTK_BOX(gtk_hbox_new(FALSE, 2));

		g_object_ref(self->sink_event_box);
		gtk_container_remove(GTK_CONTAINER(self->box), self->sink_event_box);
		gtk_box_pack_start(box, self->sink_event_box, FALSE, FALSE, 0);
		g_object_unref(self->sink_event_box);

		g_object_ref(self->source_event_box);
		gtk_container_remove(GTK_CONTAINER(self->box), self->source_event_box);
		gtk_box_pack_start(box, self->source_event_box, FALSE, FALSE, 0);
		g_object_unref(self->source_event_box);

		gtk_container_remove(GTK_CONTAINER(self), GTK_WIDGET(self->box));
		gtk_container_add(GTK_CONTAINER(self), GTK_WIDGET(box));

		gtk_widget_show(GTK_WIDGET(box));
		self->box = box;
	}
	self->orientation = orientation;

	if (self->sink_popup)
		gtk_widget_destroy(GTK_WIDGET(self->sink_popup));
}

static void pama_applet_dispose(GObject *gobject)
{
	PamaApplet *self = PAMA_APPLET(gobject);

	if (self->context)
	{
		g_object_unref(self->context);
		self->context = NULL;
	}

	G_OBJECT_CLASS(pama_applet_parent_class)->dispose(gobject);
}



static void pama_applet_update_icons(PamaApplet *self)
{
	guint32 volume;
	gboolean mute;
	gchar *description;
	gchar *tooltip;

	if (self->default_sink)
	{
		g_object_get(self->default_sink,
		             "mute", &mute,
					 "volume", &volume,
					 "description", &description,
					 NULL);

		if (mute)
		{
			gtk_image_set_from_icon_name(GTK_IMAGE(self->sink_icon), "audio-volume-muted", GTK_ICON_SIZE_MENU);
			tooltip = g_markup_printf_escaped("<b>%s</b>: %s", description, _("muted"));
		}
		else
		{
			if (volume < (PA_VOLUME_NORM/3))
				gtk_image_set_from_icon_name(GTK_IMAGE(self->sink_icon), "audio-volume-low", GTK_ICON_SIZE_MENU);
			else if (volume < (2 * PA_VOLUME_NORM / 3))
				gtk_image_set_from_icon_name(GTK_IMAGE(self->sink_icon), "audio-volume-medium", GTK_ICON_SIZE_MENU);
			else
				gtk_image_set_from_icon_name(GTK_IMAGE(self->sink_icon), "audio-volume-high", GTK_ICON_SIZE_MENU);

			tooltip = g_markup_printf_escaped("<b>%s</b>: %.0f%%", description, 100.0 * volume / PA_VOLUME_NORM);
		}
		gtk_widget_set_tooltip_markup(GTK_WIDGET(self->sink_event_box), tooltip);
		g_free(tooltip);
		g_free(description);
	}
	else
	{
		gtk_image_set_from_icon_name(GTK_IMAGE(self->sink_icon), "audio-volume-muted", GTK_ICON_SIZE_MENU);
		gtk_widget_set_tooltip_text(GTK_WIDGET(self->sink_event_box), _("No information for the default output device is available"));
	}

	if (self->default_source)
	{
		g_object_get(self->default_source,
		             "mute", &mute,
					 "volume", &volume,
					 "description", &description,
					 NULL);

		if (mute)
		{
			gtk_image_set_from_icon_name(GTK_IMAGE(self->source_icon), "audio-input-microphone-muted", GTK_ICON_SIZE_MENU);
			tooltip = g_markup_printf_escaped("<b>%s</b>: %s", description, _("muted"));
		}
		else
		{
			gtk_image_set_from_icon_name(GTK_IMAGE(self->source_icon), "audio-input-microphone", GTK_ICON_SIZE_MENU);
			tooltip = g_markup_printf_escaped("<b>%s</b>: %.0f%%", description, 100.0 * volume / PA_VOLUME_NORM);
		}
		gtk_widget_set_tooltip_markup(GTK_WIDGET(self->source_event_box), tooltip);
		g_free(tooltip);
		g_free(description);
	}
	else
	{
		gtk_image_set_from_icon_name(GTK_IMAGE(self->sink_icon), "audio-input-microphone-muted", GTK_ICON_SIZE_MENU);
		gtk_widget_set_tooltip_text(GTK_WIDGET(self->sink_event_box), _("No information for the default input device is available"));
	}

	gtk_widget_trigger_tooltip_query(GTK_WIDGET(self));
}

static gboolean pama_applet_icon_scroll(GtkWidget *event_box, GdkEventScroll *event, gpointer data)
{
	PamaApplet *self = data;
	guint volume;

	if (NULL == event)
		return FALSE;

	switch (event->direction)
	{
		case GDK_SCROLL_UP:
		case GDK_SCROLL_RIGHT:
			if (event_box == self->sink_event_box && self->default_sink)
			{
				g_object_get(self->default_sink, "volume", &volume, NULL);

				volume += VOLUME_SCROLL_STEP;
				if (volume > PA_VOLUME_NORM)
					volume = PA_VOLUME_NORM;

				pama_pulse_sink_set_volume(self->default_sink, volume);
			}
			else if (event_box == self->source_event_box && self->default_source)
			{
				g_object_get(self->default_source, "volume", &volume,
							 NULL);

				volume += VOLUME_SCROLL_STEP;
				if (volume > PA_VOLUME_NORM)
					volume = PA_VOLUME_NORM;

				pama_pulse_source_set_volume(self->default_source, volume);
			}
			break;

		case GDK_SCROLL_DOWN:
		case GDK_SCROLL_LEFT:
			if (event_box == self->sink_event_box)
			{
				g_object_get(self->default_sink, "volume", &volume, NULL);

				if (volume < VOLUME_SCROLL_STEP)
					volume = 0;
				else
					volume -= VOLUME_SCROLL_STEP;

				pama_pulse_sink_set_volume(self->default_sink, volume);
			}
			else if (event_box == self->source_event_box && self->default_source)
			{
				g_object_get(self->default_source, "volume", &volume, NULL);

				if (volume < VOLUME_SCROLL_STEP)
					volume = 0;
				else
					volume -= VOLUME_SCROLL_STEP;

				pama_pulse_source_set_volume(self->default_source, volume);
			}
			break;
	}

	return TRUE;
}
static gboolean pama_applet_icon_click (GtkWidget *event_box, GdkEventButton *event, gpointer data)
{
	PamaApplet *self = data;

	if (NULL == event)
		return FALSE;

	if (GDK_BUTTON_PRESS == event->type)
	{
		if (1 == event->button)
		{
			if (! self->context)
				return TRUE;

			if (event_box == self->sink_event_box)
			{
				if (self->sink_popup)
					gtk_widget_destroy(GTK_WIDGET(self->sink_popup));
				else
				{
					if (self->source_popup)
						gtk_widget_destroy(GTK_WIDGET(self->source_popup));

					self->sink_popup = pama_sink_popup_new(self->context);
					g_signal_connect(self->sink_popup, "destroy", G_CALLBACK(gtk_widget_destroyed),   &self->sink_popup);
					pama_popup_align_with_widget(PAMA_POPUP(self->sink_popup), self->sink_event_box, self->orientation);
					gtk_widget_show_all(GTK_WIDGET(self->sink_popup));
				}
			}
			else
			{
				if (self->source_popup)
					gtk_widget_destroy(GTK_WIDGET(self->source_popup));
				else
				{
					if (self->sink_popup)
						gtk_widget_destroy(GTK_WIDGET(self->sink_popup));

					self->source_popup = pama_source_popup_new(self->context);
					g_signal_connect(self->source_popup, "destroy", G_CALLBACK(gtk_widget_destroyed),   &self->source_popup);
					pama_popup_align_with_widget(PAMA_POPUP(self->source_popup), self->source_event_box, self->orientation);
					gtk_widget_show_all(GTK_WIDGET(self->source_popup));
				}
			}

			return TRUE;
		}
	}
	
	return FALSE;
}


static void pama_applet_show_about_dialog(BonoboUIComponent *uic, gpointer data, const char *cname)
{
	const gchar *authors[] = {
		"Vassili Geronimos <v.geronimos@gmail.com>",
		NULL
	};
	const gchar *translator_credits = _("translator-credits");
	if (0 == strcmp(translator_credits, "translator-credits"))
		translator_credits = NULL;

	gtk_show_about_dialog(NULL,
	                      "program-name", "Pulseaudio Mixer Applet",
	                      "version", PACKAGE_VERSION,
	                      "title", _("About Pulseaudio Mixer Applet"),
	                      "logo-icon-name", "multimedia-volume-control",
	                      "authors", authors,
	                      "comments", _("Per-application volume control for your GNOME panel."), 
	                      "copyright", _("© 2009 Vassili Geronimos"),
	                      "translator-credits", translator_credits,
	                      NULL);
}
