/*
 * pama-pulse-source.c: A GObject wrapper for a PulseAudio source
 * Part of PulseAudio Mixer Applet
 * Copyright © Vassili Geronimos 2009 <v.geronimos@gmail.com>
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

 
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif
#include <glib.h>
#include <glib/gi18n.h>
 
#include <string.h>
#include "pama-pulse-context.h"
#include "pama-pulse-source.h"

#define PAMA_PULSE_SOURCE_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE((obj), PAMA_TYPE_PULSE_SOURCE, PamaPulseSourcePrivate))

struct _PamaPulseSourcePrivate
{
	guint32           index;
	guint32           volume;
	guint8            channels;
	gboolean          mute;
	GString *         name;
	GString *         description;
	pa_source_flags_t flags;
	PamaPulseContext *context;
};

static void pama_pulse_source_init(PamaPulseSource *source);
static void pama_pulse_source_class_init(PamaPulseSourceClass *klass);
static void pama_pulse_source_finalize(GObject *gobject);
static void pama_pulse_source_get_property(GObject *gobject, guint property_id,       GValue *value, GParamSpec *pspec);
static void pama_pulse_source_set_property(GObject *gobject, guint property_id, const GValue *value, GParamSpec *pspec);

G_DEFINE_TYPE(PamaPulseSource, pama_pulse_source, G_TYPE_OBJECT);


enum
{
	PROP_0,

	PROP_INDEX,
	PROP_VOLUME,
	PROP_CHANNELS,
	PROP_MUTE,
	PROP_NAME,
	PROP_DESCRIPTION,
	PROP_CONTEXT
};

static void pama_pulse_source_class_init(PamaPulseSourceClass *klass)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
	GParamSpec *pspec;
	
	gobject_class->finalize = pama_pulse_source_finalize;
	gobject_class->get_property = pama_pulse_source_get_property;
	gobject_class->set_property = pama_pulse_source_set_property;
	
	pspec = g_param_spec_uint("index",
	                          "Source index",
	                          "The index assigned to this source",
	                          0,
	                          G_MAXUINT,
	                          0,
	                          G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY | G_PARAM_STATIC_STRINGS);
	g_object_class_install_property(gobject_class, PROP_INDEX, pspec);

	pspec = g_param_spec_uint("volume",
	                          "Source volume",
	                          "The raw volume of this source",
	                          0,
	                          G_MAXUINT,
	                          0,
	                          G_PARAM_READWRITE | G_PARAM_CONSTRUCT | G_PARAM_STATIC_STRINGS);
	g_object_class_install_property(gobject_class, PROP_VOLUME, pspec);

	pspec = g_param_spec_uchar("channels",
	                           "Channel count",
							   "The number of audio channels of this source",
							   0,
							   G_MAXUINT8,
							   0,
							   G_PARAM_READWRITE | G_PARAM_CONSTRUCT | G_PARAM_STATIC_STRINGS);
	g_object_class_install_property(gobject_class, PROP_CHANNELS, pspec);

	pspec = g_param_spec_boolean("mute",
	                             "Source mute flag",
	                             "Indicates whether the source has been muted",
	                             FALSE,
	                             G_PARAM_READWRITE | G_PARAM_CONSTRUCT | G_PARAM_STATIC_STRINGS);
	g_object_class_install_property(gobject_class, PROP_MUTE, pspec);

	pspec = g_param_spec_string("name",
	                            "Source identifier",
	                            "The systematic name assigned to this source.",
	                            "",
	                            G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY | G_PARAM_STATIC_STRINGS);
	g_object_class_install_property(gobject_class, PROP_NAME, pspec);

	pspec = g_param_spec_string("description",
	                            "Source description",
	                            "The descriptive name of this source.",
	                            "",
	                            G_PARAM_READWRITE | G_PARAM_CONSTRUCT | G_PARAM_STATIC_STRINGS);
	g_object_class_install_property(gobject_class, PROP_DESCRIPTION, pspec);

	pspec = g_param_spec_object("context",
	                            "Pulse context object",
	                            "The PamaPulseContext which this source belongs to.",
	                            PAMA_TYPE_PULSE_CONTEXT,
	                            G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY | G_PARAM_STATIC_STRINGS);
	g_object_class_install_property(gobject_class, PROP_CONTEXT, pspec);

	g_type_class_add_private(klass, sizeof(PamaPulseSourcePrivate));
}

static void pama_pulse_source_init(PamaPulseSource *self)
{
	PamaPulseSourcePrivate *priv;
	
	self->priv = priv = PAMA_PULSE_SOURCE_GET_PRIVATE(self);
	
	priv->name = g_string_new("");
	priv->description = g_string_new("");
}

static void pama_pulse_source_finalize(GObject *gobject)
{
	PamaPulseSource *self = PAMA_PULSE_SOURCE(gobject);

	g_string_free(self->priv->name, TRUE);
	g_string_free(self->priv->description, TRUE);

	/* Chain up to the parent class */
	G_OBJECT_CLASS (pama_pulse_source_parent_class)->finalize (gobject);
}

static void pama_pulse_source_get_property(GObject *gobject, guint property_id,       GValue *value, GParamSpec *pspec)
{
	PamaPulseSource *self = PAMA_PULSE_SOURCE(gobject);

	switch(property_id)
	{
		case PROP_INDEX:
			g_value_set_uint(value, self->priv->index);
			break;

		case PROP_VOLUME:
			g_value_set_uint(value, self->priv->volume);
			break;

		case PROP_CHANNELS:
			g_value_set_uchar(value, self->priv->channels);
			break;

		case PROP_MUTE:
			g_value_set_boolean(value, self->priv->mute);
			break;

		case PROP_NAME:
			g_value_set_string(value, self->priv->name->str);
			break;

		case PROP_DESCRIPTION:
			g_value_set_string(value, self->priv->description->str);
			break;

		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(self, property_id, pspec);
			break;
	}
}
static void pama_pulse_source_set_property(GObject *gobject, guint property_id, const GValue *value, GParamSpec *pspec)
{
	PamaPulseSource *self = PAMA_PULSE_SOURCE(gobject);

	switch(property_id)
	{
		case PROP_INDEX:
			self->priv->index = g_value_get_uint(value);
			break;

		case PROP_VOLUME:
			self->priv->volume = g_value_get_uint(value);
			break;

		case PROP_CHANNELS:
			self->priv->channels = g_value_get_uchar(value);
			break;

		case PROP_MUTE:
			self->priv->mute = g_value_get_boolean(value);
			break;

		case PROP_NAME:
			g_string_assign(self->priv->name, g_value_get_string(value));
			break;

		case PROP_DESCRIPTION:
			g_string_assign(self->priv->description, g_value_get_string(value));
			break;
			
		case PROP_CONTEXT:
			self->priv->context = g_value_get_object(value);
			break;

		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(self, property_id, pspec);
			break;
	}
}

gint pama_pulse_source_compare_to_index(gconstpointer a, gconstpointer b)
{
	const PamaPulseSource *A = a;
	const guint32 *B = b;
	
	return A->priv->index - *B;
}
gint pama_pulse_source_compare_to_name(gconstpointer a, gconstpointer b)
{
	const PamaPulseSource *A = a;
	const char *B = b;
	
	return strcmp(A->priv->name->str, B);
}
gint pama_pulse_source_compare_by_description(gconstpointer a, gconstpointer b)
{
	const PamaPulseSource *A = a;
	const PamaPulseSource *B = b;
	
	return strcmp(A->priv->description->str, B->priv->description->str);
}


void pama_pulse_source_set_mute(PamaPulseSource *self, gboolean mute)
{
	pa_context *c;
	g_object_get(self->priv->context,
	             "context", &c, 
				 NULL);

	pa_operation *o = pa_context_set_source_mute_by_index(c, self->priv->index, mute, NULL, NULL);
	if (o)
		pa_operation_unref(o);
}

void pama_pulse_source_set_volume(PamaPulseSource *self, const guint32 volume)
{
	pa_context *c;
	pa_cvolume cvolume;
	g_object_get(self->priv->context,
	             "context", &c, 
				 NULL);

	pa_cvolume_set(&cvolume, self->priv->channels, volume);
	pa_operation *o = pa_context_set_source_volume_by_index(c, self->priv->index, &cvolume, NULL, NULL);
	if (o)
		pa_operation_unref(o);
}

void pama_pulse_source_set_as_default(PamaPulseSource *self)
{
	pa_context   *c;
	pa_operation *o;
	g_object_get(self->priv->context,
	             "context", &c,
	             NULL);

	o = pa_context_set_default_source(c, self->priv->name->str, NULL, NULL);
	if (o)
		pa_operation_unref(o);
}

