/*
 * pama-sink-input-widget.c: A widget to display and manipulate a sink input
 * Part of PulseAudio Mixer Applet
 * Copyright © Vassili Geronimos 2009 <v.geronimos@gmail.com>
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif
#include <glib.h>
#include <glib/gi18n.h>
 
#include "pama-sink-input-widget.h"
#include "widget-settings.h"

static void     pama_sink_input_widget_class_init(PamaSinkInputWidgetClass *klass);
static void     pama_sink_input_widget_init(PamaSinkInputWidget *self);
static void     pama_sink_input_widget_set_property(GObject *gobject, guint property_id, const GValue *value, GParamSpec *pspec);
static GObject* pama_sink_input_widget_constructor(GType gtype, guint n_properties, GObjectConstructParam *properties);
static void     pama_sink_input_widget_dispose(GObject *gobject);
static void     pama_sink_input_widget_weak_ref_notify(gpointer data, GObject *where_the_object_was);

static void     pama_sink_input_widget_sink_input_notify(GObject *gobject, GParamSpec *pspec, PamaSinkInputWidget *self);

static void     pama_sink_input_widget_update_values(PamaSinkInputWidget *self);
static void     pama_sink_input_widget_mute_toggled(GtkToggleButton *togglebutton, PamaSinkInputWidget *self);
static void     pama_sink_input_widget_volume_changed(GtkRange *range, PamaSinkInputWidget *self);

static void     pama_sink_input_widget_sink_button_clicked(GtkButton *button, PamaSinkInputWidget *self);
static void     pama_sink_input_widget_sink_menu_hidden(GtkWidget *menu_window, PamaSinkInputWidget *self);
static void     pama_sink_input_widget_sink_menu_item_toggled(GtkCheckMenuItem *item, PamaSinkInputWidget *self);

G_DEFINE_TYPE(PamaSinkInputWidget, pama_sink_input_widget, GTK_TYPE_HBOX);


enum 
{
	PROP_0,

	PROP_SINK_INPUT,
	PROP_CONTEXT
};
enum
{
	REORDER_REQUEST_SIGNAL,
	LAST_SIGNAL
};
static guint widget_signals[LAST_SIGNAL] = {0,};

static void pama_sink_input_widget_class_init(PamaSinkInputWidgetClass *klass)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
	GParamSpec *pspec;

	gobject_class->set_property = pama_sink_input_widget_set_property;
	gobject_class->constructor  = pama_sink_input_widget_constructor;
	gobject_class->dispose      = pama_sink_input_widget_dispose;

	pspec = g_param_spec_object("sink-input",
	                            "Pulse sink-input",
	                            "The PamaPulseSinkInput that is to be controlled.",
	                            PAMA_TYPE_PULSE_SINK_INPUT,
	                            G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY | G_PARAM_STATIC_STRINGS);
	g_object_class_install_property(gobject_class, PROP_SINK_INPUT, pspec);

	pspec = g_param_spec_object("context",
	                            "Pulse context",
	                            "The PamaPulseContext that is managing the current connection.",
	                            PAMA_TYPE_PULSE_CONTEXT,
	                            G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY | G_PARAM_STATIC_STRINGS);
	g_object_class_install_property(gobject_class, PROP_CONTEXT, pspec);

	widget_signals[REORDER_REQUEST_SIGNAL] =
		g_signal_new("reorder-request",
		             G_TYPE_FROM_CLASS(gobject_class),
		             G_SIGNAL_RUN_LAST,
		             0,
		             NULL,
		             NULL,
		             g_cclosure_marshal_VOID__VOID,
		             G_TYPE_NONE,
		             0);
}

static void pama_sink_input_widget_init(PamaSinkInputWidget *self)
{
}

static void pama_sink_input_widget_set_property(GObject *gobject, guint property_id, const GValue *value, GParamSpec *pspec)
{
	PamaSinkInputWidget *self = PAMA_SINK_INPUT_WIDGET(gobject);

	switch(property_id)
	{
		case PROP_SINK_INPUT:
			self->sink_input = g_value_get_object(value);
			g_object_weak_ref(G_OBJECT(self->sink_input), pama_sink_input_widget_weak_ref_notify, self);
			break;

		case PROP_CONTEXT:
			self->context = g_value_get_object(value);
			g_object_weak_ref(G_OBJECT(self->context), pama_sink_input_widget_weak_ref_notify, self);
			break;

		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(gobject, property_id, pspec);
			break;
	}
}
static GObject* pama_sink_input_widget_constructor(GType gtype, guint n_properties, GObjectConstructParam *properties)
{
	GtkWidget *name, *alignment;
	GtkWidget *inner_box, *volume, *value, *mute, *sink_button;

	GObject *gobject = G_OBJECT_CLASS(pama_sink_input_widget_parent_class)->constructor(gtype, n_properties, properties);
	PamaSinkInputWidget *self = PAMA_SINK_INPUT_WIDGET(gobject);

	if (NULL == self->sink_input)
		g_error("An attempt was made to create a sink_input widget with no sink_input.");
	if (NULL == self->context)
		g_error("An attempt was made to create a sink_input widget with no context.");

	name = g_object_new(GTK_TYPE_LABEL,
	                    "ellipsize", PANGO_ELLIPSIZE_MIDDLE,
	                    "wrap", FALSE,
	                    "width-chars", WIDGET_NAME_WIDTH_IN_CHARS,
	                    "xalign", 0.0f,
	                    NULL);
	gtk_box_pack_start(GTK_BOX(self), name, FALSE, FALSE, 0);
	self->name = name;

	alignment = gtk_alignment_new(0, 0.5, 0, 0);
	gtk_box_pack_start(GTK_BOX(self), alignment, FALSE, FALSE, 0);

	inner_box = gtk_hbox_new(FALSE, 6);
	gtk_container_add(GTK_CONTAINER(alignment), inner_box);

	volume = gtk_hscale_new_with_range(0, WIDGET_VOLUME_SLIDER_DB_RANGE, 1.5); /* for scroll steps of 5% */
	gtk_scale_set_draw_value   (GTK_SCALE(volume), FALSE);
	gtk_widget_set_size_request(volume, WIDGET_VOLUME_SLIDER_WIDTH, -1);
	gtk_box_pack_start(GTK_BOX(inner_box), volume, FALSE, FALSE, 0);
	self->volume = volume;

	value = g_object_new(GTK_TYPE_LABEL,
	                     "width-chars", WIDGET_VALUE_WIDTH_IN_CHARS,
	                     "xalign", 1.0f,
	                     NULL);
	gtk_box_pack_start(GTK_BOX(inner_box), value, FALSE, FALSE, 0);
	self->value = value;

	mute = gtk_check_button_new();
	gtk_toggle_button_set_mode(GTK_TOGGLE_BUTTON(mute), FALSE);
	gtk_widget_set_tooltip_text(GTK_WIDGET(mute), _("Mute this application's audio output"));
	gtk_container_add(GTK_CONTAINER(mute), gtk_image_new_from_icon_name("audio-volume-muted", GTK_ICON_SIZE_MENU));
	gtk_box_pack_start(GTK_BOX(inner_box), mute, FALSE, FALSE, 0);
	self->mute = mute;

	sink_button = gtk_button_new();
	gtk_widget_set_tooltip_text(GTK_WIDGET(sink_button), _("Select which output device to use for this application"));
	gtk_container_add(GTK_CONTAINER(sink_button), gtk_image_new_from_icon_name("audio-card", GTK_ICON_SIZE_MENU));
	gtk_box_pack_start(GTK_BOX(inner_box), sink_button, FALSE, FALSE, 0);
	self->sink_button = sink_button;

	pama_sink_input_widget_update_values(self);

	g_signal_connect(volume,      "value-changed", G_CALLBACK(pama_sink_input_widget_volume_changed),      self);
	g_signal_connect(mute,        "toggled",       G_CALLBACK(pama_sink_input_widget_mute_toggled),        self);
	g_signal_connect(sink_button, "clicked",       G_CALLBACK(pama_sink_input_widget_sink_button_clicked), self);

	self->sink_input_notify_handler_id = g_signal_connect(self->sink_input, "notify::volume", G_CALLBACK(pama_sink_input_widget_sink_input_notify), self);

	return gobject;
}
static void pama_sink_input_widget_dispose(GObject *gobject)
{
	PamaSinkInputWidget *self = PAMA_SINK_INPUT_WIDGET(gobject);

	if (self->sink_input)
	{
		if (self->sink_input_notify_handler_id)
		{
			g_signal_handler_disconnect(self->sink_input, self->sink_input_notify_handler_id);
			self->sink_input_notify_handler_id = 0;
		}

		g_object_weak_unref(G_OBJECT(self->sink_input), pama_sink_input_widget_weak_ref_notify, self);
		self->sink_input = NULL;
	}

	if (self->context)
	{
		g_object_weak_unref(G_OBJECT(self->context), pama_sink_input_widget_weak_ref_notify, self);
		self->context = NULL;
	}

	if (self->sink_menu)
	{
		g_object_unref(self->sink_menu);
		self->sink_menu = NULL;
	}

	G_OBJECT_CLASS(pama_sink_input_widget_parent_class)->dispose(gobject);
}
static void pama_sink_input_widget_weak_ref_notify(gpointer data, GObject *where_the_object_was)
{
	PamaSinkInputWidget *self = data;

	if ((GObject *)self->sink_input == where_the_object_was)
		self->sink_input = NULL;

	if ((GObject *)self->context    == where_the_object_was)
		self->context = NULL;

	/* widget is no longer usable without these items. */
	gtk_object_destroy(GTK_OBJECT(self));
}


static void pama_sink_input_widget_sink_input_notify(GObject *gobject, GParamSpec *pspec, PamaSinkInputWidget *self)
{
	pama_sink_input_widget_update_values(self);
	g_signal_emit(self, widget_signals[REORDER_REQUEST_SIGNAL], 0);
}

static void pama_sink_input_widget_update_values(PamaSinkInputWidget *self)
{
	gchar *temp;
	gchar *stream_name, *client_name;
	guint            volume;
	gboolean         mute;
	PamaPulseClient *client;

	double volume_dB;

	self->updating = TRUE;

	g_object_get(self->sink_input,
	             "volume", &volume,
	             "mute",   &mute,
	             "client", &client,
	             "name",   &stream_name,
	             NULL);

	g_object_get(client, "name", &client_name, NULL);

	temp = g_markup_printf_escaped("<b>%s</b>\n%s", client_name, stream_name);
	gtk_label_set_markup(GTK_LABEL(self->name), temp);
	gtk_widget_set_tooltip_markup(GTK_WIDGET(self->name), temp);
	g_free(temp);
	g_free(stream_name);
	g_free(client_name);

	volume_dB = pa_sw_volume_to_dB(volume);
	if (isinf(volume_dB))
		gtk_label_set_text(GTK_LABEL(self->value), "-∞dB");
	else
	{
		temp = g_strdup_printf("%.1fdB", volume_dB);
		gtk_label_set_text(GTK_LABEL(self->value), temp);
		g_free(temp);
	}

	gtk_range_set_value(GTK_RANGE(self->volume), volume_dB + WIDGET_VOLUME_SLIDER_DB_RANGE);
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(self->mute), mute);

	gtk_widget_set_sensitive(self->volume, !mute);
	gtk_widget_set_sensitive(self->value,  !mute);

	self->updating = FALSE;
}

static void pama_sink_input_widget_mute_toggled   (GtkToggleButton *togglebutton, PamaSinkInputWidget *self)
{
	if (self->updating)
		return;

	pama_pulse_sink_input_set_mute(self->sink_input, gtk_toggle_button_get_active(togglebutton));
}
static void pama_sink_input_widget_volume_changed (GtkRange *range, PamaSinkInputWidget *self)
{
	if (self->updating)
		return;

	pama_pulse_sink_input_set_volume(self->sink_input, pa_sw_volume_from_dB(gtk_range_get_value(range) - WIDGET_VOLUME_SLIDER_DB_RANGE));
}

static void pama_sink_input_widget_sink_button_clicked(GtkButton *button, PamaSinkInputWidget *self)
{
	GtkWidget *item, *last = NULL;
	GSList *sinks, *iter;
	PamaPulseSink *sink, *current_sink;
	gchar *sink_name;

	if (!self->sink_menu)
	{
		self->sink_menu = gtk_menu_new();
		g_object_ref_sink(self->sink_menu);
		g_signal_connect(gtk_widget_get_toplevel(self->sink_menu), "hide", G_CALLBACK(pama_sink_input_widget_sink_menu_hidden), self);

		sinks = g_slist_copy(pama_pulse_context_get_sinks(self->context));
		sinks = g_slist_sort(sinks, pama_pulse_sink_compare_by_description);

		g_object_get(self->sink_input, "sink", &current_sink, NULL);

		for (iter = sinks; iter; iter = iter->next)
		{
			sink = PAMA_PULSE_SINK(iter->data);
			g_object_get(sink, "description", &sink_name, NULL);

			if (last)
				item = gtk_radio_menu_item_new_with_label_from_widget(GTK_RADIO_MENU_ITEM(last), sink_name);
			else
				item = gtk_radio_menu_item_new_with_label(NULL, sink_name);

			g_object_set(item,
			             "active", (current_sink == sink),
			             "user-data", sink,
			             NULL);
			last = item;
			g_free(sink_name);

			g_signal_connect(item, "toggled", G_CALLBACK(pama_sink_input_widget_sink_menu_item_toggled), self);
			gtk_menu_shell_append(GTK_MENU_SHELL(self->sink_menu), item);
			gtk_widget_show_all(item);
		}

		g_slist_free(sinks);

		gtk_menu_popup(GTK_MENU(self->sink_menu), NULL, NULL, NULL, NULL, 0, gtk_get_current_event_time());
	}
}
static void pama_sink_input_widget_sink_menu_hidden(GtkWidget *menu_window, PamaSinkInputWidget *self)
{
	g_object_unref(self->sink_menu);
	self->sink_menu = NULL;
}
static void pama_sink_input_widget_sink_menu_item_toggled(GtkCheckMenuItem *item, PamaSinkInputWidget *self)
{
	PamaPulseSink *sink;

	if (!gtk_check_menu_item_get_active(item))
		return;

	g_object_get(item, "user-data", &sink, NULL);

	pama_pulse_sink_input_set_sink(self->sink_input, sink);
}

gint pama_sink_input_widget_compare(gconstpointer a, gconstpointer b)
{
	PamaSinkInputWidget *A = PAMA_SINK_INPUT_WIDGET(a);
	PamaSinkInputWidget *B = PAMA_SINK_INPUT_WIDGET(b);

	return pama_pulse_sink_input_compare_by_name(A->sink_input, B->sink_input);
}
