/*
 * pama-sink-popup.c: A popup window that displays all sinks and sink inputs
 * Part of PulseAudio Mixer Applet
 * Copyright © Vassili Geronimos 2009 <v.geronimos@gmail.com>
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

 
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif
#include <glib.h>
#include <glib/gi18n.h>
 
#include "pama-sink-popup.h"
#include "pama-sink-widget.h"
#include "pama-sink-input-widget.h"

static void     pama_sink_popup_set_property(GObject *gobject, guint property_id, const GValue *value, GParamSpec *pspec);
static GObject* pama_sink_popup_constructor(GType gtype, guint n_properties, GObjectConstructParam *properties);
static void     pama_sink_popup_dispose(GObject *gobject);
static void     pama_sink_popup_weak_ref_notify(gpointer data, GObject *where_the_object_was);

static void     pama_sink_popup_sink_added(PamaPulseContext *context, guint index, gpointer data);
static void     pama_sink_popup_sink_input_added(PamaPulseContext *context, guint index, gpointer data);

static void     pama_sink_popup_reorder_sinks      (PamaSinkWidget      *widget, gpointer data);
static void     pama_sink_popup_reorder_sink_inputs(PamaSinkInputWidget *widget, gpointer data);

static void     pama_sink_popup_add_sink      (PamaSinkPopup *self, PamaPulseSink      *sink);
static void     pama_sink_popup_add_sink_input(PamaSinkPopup *self, PamaPulseSinkInput *sink_input);

G_DEFINE_TYPE(PamaSinkPopup, pama_sink_popup, PAMA_TYPE_POPUP);

enum
{
	PROP_0,

	PROP_CONTEXT
};

static void pama_sink_popup_class_init(PamaSinkPopupClass *klass)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
	GParamSpec *pspec;

	gobject_class->set_property = pama_sink_popup_set_property;
	gobject_class->constructor  = pama_sink_popup_constructor;
	gobject_class->dispose      = pama_sink_popup_dispose;

	pspec = g_param_spec_object("context",
	                            "Pulse context object",
	                            "The PamaPulseContext which this sink belongs to.",
	                            PAMA_TYPE_PULSE_CONTEXT,
	                            G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY | G_PARAM_STATIC_STRINGS);
	g_object_class_install_property(gobject_class, PROP_CONTEXT, pspec);
}

static void pama_sink_popup_init(PamaSinkPopup *popup)
{
	gchar *markup;
	GtkWidget *frame, *main_box;
	GtkWidget *sink_frame,   *sink_align,   *sink_box;
	GtkWidget *stream_frame, *stream_align, *stream_box;

	frame = gtk_frame_new(NULL);
	gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_OUT);
	gtk_container_add(GTK_CONTAINER(popup), frame);

	main_box = gtk_vbox_new(FALSE, 18);
	gtk_container_set_border_width(GTK_CONTAINER(main_box), 6);
	gtk_container_add(GTK_CONTAINER(frame), main_box);


	sink_frame = gtk_frame_new("");
	markup = g_markup_printf_escaped("<b>%s</b>", _("Playback Devices"));
	gtk_label_set_markup(GTK_LABEL(gtk_frame_get_label_widget(GTK_FRAME(sink_frame))), markup);
	gtk_frame_set_shadow_type(GTK_FRAME(sink_frame), GTK_SHADOW_NONE);
	gtk_box_pack_start(GTK_BOX(main_box), sink_frame, FALSE, FALSE, 0);
	g_free(markup);
	
	sink_align = gtk_alignment_new(0,0,1,0);
	gtk_alignment_set_padding(GTK_ALIGNMENT(sink_align), 6, 0, 12, 0);
	gtk_container_add(GTK_CONTAINER(sink_frame), sink_align);

	sink_box = gtk_vbox_new(FALSE, 6);
	gtk_container_add(GTK_CONTAINER(sink_align), sink_box);
	popup->sink_box = GTK_BOX(sink_box);

	
	stream_frame = gtk_frame_new("");
	markup = g_markup_printf_escaped("<b>%s</b>", _("Applications"));
	gtk_label_set_markup(GTK_LABEL(gtk_frame_get_label_widget(GTK_FRAME(stream_frame))), markup);
	gtk_frame_set_shadow_type(GTK_FRAME(stream_frame), GTK_SHADOW_NONE);
	gtk_box_pack_start(GTK_BOX(main_box), stream_frame, FALSE, FALSE, 0);
	g_free(markup);
	
	stream_align = gtk_alignment_new(0,0,1,0);
	gtk_alignment_set_padding(GTK_ALIGNMENT(stream_align), 6, 0, 12, 0);
	gtk_container_add(GTK_CONTAINER(stream_frame), stream_align);

	stream_box = gtk_vbox_new(FALSE, 6);
	gtk_container_add(GTK_CONTAINER(stream_align), stream_box);
	popup->stream_box = GTK_BOX(stream_box);
}

static GObject* pama_sink_popup_constructor(GType gtype, guint n_properties, GObjectConstructParam *properties)
{
	GObject *gobject = G_OBJECT_CLASS(pama_sink_popup_parent_class)->constructor(gtype, n_properties, properties);
	PamaSinkPopup *self = PAMA_SINK_POPUP(gobject);
	GSList *iter;

	if (NULL == self->context)
		g_error("An attempt was made to construct a PamaSinkPopup without providing a valid PamaPulseContext.");

	for (iter = pama_pulse_context_get_sinks(self->context); iter; iter = iter->next)
	{
		PamaPulseSink *sink = PAMA_PULSE_SINK(iter->data);

		pama_sink_popup_add_sink(self, sink);
	}

	for (iter = pama_pulse_context_get_sink_inputs(self->context); iter; iter = iter->next)
	{
		PamaPulseSinkInput *sink_input = PAMA_PULSE_SINK_INPUT(iter->data);

		pama_sink_popup_add_sink_input(self, sink_input);
	}

	pama_sink_popup_reorder_sinks(NULL, self);
	self->sink_added_handler_id       = g_signal_connect(self->context, "sink-added",       G_CALLBACK(pama_sink_popup_sink_added),       self);
	self->sink_input_added_handler_id = g_signal_connect(self->context, "sink-input-added", G_CALLBACK(pama_sink_popup_sink_input_added), self);

	gtk_widget_show_all(GTK_BIN(self)->child);

	return gobject;
}
static void pama_sink_popup_dispose(GObject *gobject)
{
	PamaSinkPopup *self = PAMA_SINK_POPUP(gobject);

	if (self->context)
	{
		if (self->sink_added_handler_id)
		{
			g_signal_handler_disconnect(self->context, self->sink_added_handler_id);
			self->sink_added_handler_id = 0;
		}

		if (self->sink_input_added_handler_id)
		{
			g_signal_handler_disconnect(self->context, self->sink_input_added_handler_id);
			self->sink_input_added_handler_id = 0;
		}

		g_object_weak_unref(G_OBJECT(self->context), pama_sink_popup_weak_ref_notify, self);
		self->context = NULL;
	}

	G_OBJECT_CLASS(pama_sink_popup_parent_class)->dispose(gobject);
}
static void pama_sink_popup_weak_ref_notify(gpointer data, GObject *where_the_object_was)
{
	PamaSinkPopup *self = PAMA_SINK_POPUP(data);
	if ((GObject *)self->context == where_the_object_was)
	{
		self->context = NULL;
	}

	gtk_object_destroy(GTK_OBJECT(self));
}

static void pama_sink_popup_set_property(GObject *gobject, guint property_id, const GValue *value, GParamSpec *pspec)
{
	PamaSinkPopup *self = PAMA_SINK_POPUP(gobject);

	switch(property_id)
	{
		case PROP_CONTEXT:
			self->context = g_value_get_object(value);
			g_object_weak_ref(G_OBJECT(self->context), pama_sink_popup_weak_ref_notify, self);
			break;

		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(gobject, property_id, pspec);
			break;
	}
}

static void  pama_sink_popup_add_sink(PamaSinkPopup *self, PamaPulseSink *sink)
{
	GtkBoxChild *box_child;
	GtkWidget *group = NULL;
	GtkWidget *sink_widget;
	
	if (self->sink_box->children)
	{
		box_child = self->sink_box->children->data;
		group = box_child->widget;
	}
	
	sink_widget = 
		g_object_new(PAMA_TYPE_SINK_WIDGET,
		             "context", self->context, 
		             "sink", sink,
		             "group", group,
		             NULL);

	g_signal_connect(sink_widget, "reorder-request", G_CALLBACK(pama_sink_popup_reorder_sinks), self);
	
	gtk_box_pack_start(GTK_BOX(self->sink_box), sink_widget, FALSE, FALSE, 0);
	pama_sink_popup_reorder_sinks(NULL, self);
	gtk_widget_show_all(sink_widget);
}
static void pama_sink_popup_add_sink_input(PamaSinkPopup *self, PamaPulseSinkInput *sink_input)
{
	GtkBoxChild *box_child;
	GtkWidget *group = NULL;
	GtkWidget *sink_input_widget;
	
	if (self->stream_box->children)
	{
		box_child = self->stream_box->children->data;
		group = box_child->widget;
	}
	
	sink_input_widget = 
		g_object_new(PAMA_TYPE_SINK_INPUT_WIDGET,
		             "context", self->context, 
		             "sink-input", sink_input,
		             NULL);

	g_signal_connect(sink_input_widget, "reorder-request", G_CALLBACK(pama_sink_popup_reorder_sink_inputs), self);
	
	gtk_box_pack_start(GTK_BOX(self->stream_box), sink_input_widget, FALSE, FALSE, 0);
	pama_sink_popup_reorder_sink_inputs(NULL, self);
	gtk_widget_show_all(sink_input_widget);
}

static void pama_sink_popup_sink_added(PamaPulseContext *context, guint index, gpointer data)
{
	PamaSinkPopup *self = PAMA_SINK_POPUP(data);
	PamaPulseSink *sink = pama_pulse_context_get_sink_by_index(context, index);

	pama_sink_popup_add_sink(self, sink);
	pama_sink_popup_reorder_sinks(NULL, self);
}
static void pama_sink_popup_sink_input_added(PamaPulseContext *context, guint index, gpointer data)
{
	PamaSinkPopup      *self       = PAMA_SINK_POPUP(data);
	PamaPulseSinkInput *sink_input = pama_pulse_context_get_sink_input_by_index(context, index);

	pama_sink_popup_add_sink_input(self, sink_input);
	pama_sink_popup_reorder_sink_inputs(NULL, self);
}

static gint pama_sink_popup_reorder_sinks__compare_sinks(gconstpointer a, gconstpointer b)
{
	const GtkBoxChild *A = a, *B = b;

	return pama_sink_widget_compare(PAMA_SINK_WIDGET(A->widget), PAMA_SINK_WIDGET(B->widget));
}
static void pama_sink_popup_reorder_sinks(PamaSinkWidget *widget, gpointer data)
{
	PamaSinkPopup *self = PAMA_SINK_POPUP(data);
	GList *children = g_list_copy(GTK_BOX(self->sink_box)->children);
	GList *iter;
	guint position;

	children = g_list_sort(children, pama_sink_popup_reorder_sinks__compare_sinks);

	for (iter = children, position = 0; iter; iter = iter->next, position++)
	{
		GtkBoxChild *child = iter->data;
		gtk_box_reorder_child(self->sink_box, child->widget, position);
	}

	g_list_free(children);
}
static gint pama_sink_popup_reorder_sink_inputs__compare_sink_inputs(gconstpointer a, gconstpointer b)
{
	const GtkBoxChild *A = a, *B = b;

	return pama_sink_input_widget_compare(PAMA_SINK_INPUT_WIDGET(A->widget), PAMA_SINK_INPUT_WIDGET(B->widget));
}
static void pama_sink_popup_reorder_sink_inputs(PamaSinkInputWidget *widget, gpointer data)
{
	PamaSinkPopup *self = PAMA_SINK_POPUP(data);
	GList *children = g_list_copy(GTK_BOX(self->stream_box)->children);
	GList *iter;
	guint position;

	children = g_list_sort(children, pama_sink_popup_reorder_sink_inputs__compare_sink_inputs);

	for (iter = children, position = 0; iter; iter = iter->next, position++)
	{
		GtkBoxChild *child = iter->data;
		gtk_box_reorder_child(self->stream_box, child->widget, position);
	}

	g_list_free(children);
}

PamaSinkPopup* pama_sink_popup_new(PamaPulseContext *context)
{
	return g_object_new(PAMA_TYPE_SINK_POPUP,
	                    "type", GTK_WINDOW_TOPLEVEL,
	                    "type-hint", GDK_WINDOW_TYPE_HINT_DOCK,
	                    "decorated", FALSE,
	                    "resizable", FALSE,
	                    "skip-pager-hint", TRUE,
	                    "skip-taskbar-hint", TRUE,
	                    "title", "Playback Volume Widget",
	                    "icon-name", "multimedia-volume-widget",
	                    "context", context,
	                    NULL);
}
