/*
 * pama-source-output-widget.c: A widget to display and manipulate a source output
 * Part of PulseAudio Mixer Applet
 * Copyright © Vassili Geronimos 2009 <v.geronimos@gmail.com>
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif
#include <glib.h>
#include <glib/gi18n.h>
 
#include "pama-source-output-widget.h"
#include "widget-settings.h"

static void     pama_source_output_widget_class_init(PamaSourceOutputWidgetClass *klass);
static void     pama_source_output_widget_init(PamaSourceOutputWidget *self);
static void     pama_source_output_widget_set_property(GObject *gobject, guint property_id, const GValue *value, GParamSpec *pspec);
static GObject* pama_source_output_widget_constructor(GType gtype, guint n_properties, GObjectConstructParam *properties);
static void     pama_source_output_widget_dispose(GObject *gobject);
static void     pama_source_output_widget_weak_ref_notify(gpointer data, GObject *where_the_object_was);

static void     pama_source_output_widget_source_output_notify(GObject *gobject, GParamSpec *pspec, PamaSourceOutputWidget *self);

static void     pama_source_output_widget_update_values(PamaSourceOutputWidget *self);

static void     pama_source_output_widget_source_button_clicked(GtkButton *button, PamaSourceOutputWidget *self);
static void     pama_source_output_widget_source_menu_hidden(GtkWidget *menu_window, PamaSourceOutputWidget *self);
static void     pama_source_output_widget_source_menu_item_toggled(GtkCheckMenuItem *item, PamaSourceOutputWidget *self);

G_DEFINE_TYPE(PamaSourceOutputWidget, pama_source_output_widget, GTK_TYPE_HBOX);


enum 
{
	PROP_0,

	PROP_SOURCE_OUTPUT,
	PROP_CONTEXT
};
enum
{
	REORDER_REQUEST_SIGNAL,
	LAST_SIGNAL
};
static guint widget_signals[LAST_SIGNAL] = {0,};

static void pama_source_output_widget_class_init(PamaSourceOutputWidgetClass *klass)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
	GParamSpec *pspec;

	gobject_class->set_property = pama_source_output_widget_set_property;
	gobject_class->constructor  = pama_source_output_widget_constructor;
	gobject_class->dispose      = pama_source_output_widget_dispose;

	pspec = g_param_spec_object("source-output",
	                            "Pulse source-output",
	                            "The PamaPulseSourceOutput that is to be controlled.",
	                            PAMA_TYPE_PULSE_SOURCE_OUTPUT,
	                            G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY | G_PARAM_STATIC_STRINGS);
	g_object_class_install_property(gobject_class, PROP_SOURCE_OUTPUT, pspec);

	pspec = g_param_spec_object("context",
	                            "Pulse context",
	                            "The PamaPulseContext that is managing the current connection.",
	                            PAMA_TYPE_PULSE_CONTEXT,
	                            G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY | G_PARAM_STATIC_STRINGS);
	g_object_class_install_property(gobject_class, PROP_CONTEXT, pspec);

	widget_signals[REORDER_REQUEST_SIGNAL] =
		g_signal_new("reorder-request",
		             G_TYPE_FROM_CLASS(gobject_class),
		             G_SIGNAL_RUN_LAST,
		             0,
		             NULL,
		             NULL,
		             g_cclosure_marshal_VOID__VOID,
		             G_TYPE_NONE,
		             0);
}

static void pama_source_output_widget_init(PamaSourceOutputWidget *self)
{
}

static void pama_source_output_widget_set_property(GObject *gobject, guint property_id, const GValue *value, GParamSpec *pspec)
{
	PamaSourceOutputWidget *self = PAMA_SOURCE_OUTPUT_WIDGET(gobject);

	switch(property_id)
	{
		case PROP_SOURCE_OUTPUT:
			self->source_output = g_value_get_object(value);
			g_object_weak_ref(G_OBJECT(self->source_output), pama_source_output_widget_weak_ref_notify, self);
			break;

		case PROP_CONTEXT:
			self->context = g_value_get_object(value);
			g_object_weak_ref(G_OBJECT(self->context), pama_source_output_widget_weak_ref_notify, self);
			break;

		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(gobject, property_id, pspec);
			break;
	}
}
static GObject* pama_source_output_widget_constructor(GType gtype, guint n_properties, GObjectConstructParam *properties)
{
	GtkWidget *name, *alignment;
	GtkWidget *inner_box, *source_button;

	GObject *gobject = G_OBJECT_CLASS(pama_source_output_widget_parent_class)->constructor(gtype, n_properties, properties);
	PamaSourceOutputWidget *self = PAMA_SOURCE_OUTPUT_WIDGET(gobject);

	if (NULL == self->source_output)
		g_error("An attempt was made to create a source_output widget with no source_output.");
	if (NULL == self->context)
		g_error("An attempt was made to create a source_output widget with no context.");

	name = g_object_new(GTK_TYPE_LABEL,
	                    "ellipsize", PANGO_ELLIPSIZE_MIDDLE,
	                    "wrap", FALSE,
	                    "xalign", 0.0f,
	                    NULL);
	gtk_box_pack_start(GTK_BOX(self), name, TRUE, TRUE, 0);
	self->name = name;

	alignment = gtk_alignment_new(0, 0.5, 0, 0);
	gtk_box_pack_end(GTK_BOX(self), alignment, FALSE, FALSE, 0);

	source_button = gtk_button_new();
	gtk_widget_set_tooltip_text(GTK_WIDGET(source_button), _("Select which input device to use for this application"));
	gtk_container_add(GTK_CONTAINER(source_button), gtk_image_new_from_icon_name("audio-card", GTK_ICON_SIZE_MENU));
	gtk_container_add(GTK_CONTAINER(alignment), source_button);
	self->source_button = source_button;

	pama_source_output_widget_update_values(self);

	g_signal_connect(source_button, "clicked", G_CALLBACK(pama_source_output_widget_source_button_clicked), self);

	self->source_output_notify_handler_id = g_signal_connect(self->source_output, "notify::source", G_CALLBACK(pama_source_output_widget_source_output_notify), self);

	return gobject;
}
static void pama_source_output_widget_dispose(GObject *gobject)
{
	PamaSourceOutputWidget *self = PAMA_SOURCE_OUTPUT_WIDGET(gobject);

	if (self->source_output)
	{
		if (self->source_output_notify_handler_id)
		{
			g_signal_handler_disconnect(self->source_output, self->source_output_notify_handler_id);
			self->source_output_notify_handler_id = 0;
		}

		g_object_weak_unref(G_OBJECT(self->source_output), pama_source_output_widget_weak_ref_notify, self);
		self->source_output = NULL;
	}

	if (self->context)
	{
		g_object_weak_unref(G_OBJECT(self->context), pama_source_output_widget_weak_ref_notify, self);
		self->context = NULL;
	}

	if (self->source_menu)
	{
		g_object_unref(self->source_menu);
		self->source_menu = NULL;
	}

	G_OBJECT_CLASS(pama_source_output_widget_parent_class)->dispose(gobject);
}
static void pama_source_output_widget_weak_ref_notify(gpointer data, GObject *where_the_object_was)
{
	PamaSourceOutputWidget *self = data;

	if ((GObject *)self->source_output    == where_the_object_was)
		self->source_output = NULL;

	if ((GObject *)self->context == where_the_object_was)
		self->context = NULL;

	/* widget is no longer usable without these items. */
	gtk_object_destroy(GTK_OBJECT(self));
}


static void pama_source_output_widget_source_output_notify(GObject *gobject, GParamSpec *pspec, PamaSourceOutputWidget *self)
{
	pama_source_output_widget_update_values(self);
	g_signal_emit(self, widget_signals[REORDER_REQUEST_SIGNAL], 0);
}

static void pama_source_output_widget_update_values(PamaSourceOutputWidget *self)
{
	gchar *temp;
	gchar *stream_name, *client_name;
	PamaPulseClient *client;

	self->updating = TRUE;

	g_object_get(self->source_output,
	             "client", &client,
	             "name",   &stream_name,
	             NULL);

	g_object_get(client, "name", &client_name, NULL);

	temp = g_markup_printf_escaped("<b>%s</b>\n%s", client_name, stream_name);
	gtk_label_set_markup(GTK_LABEL(self->name), temp);
	gtk_widget_set_tooltip_markup(GTK_WIDGET(self->name), temp);
	g_free(temp);
	g_free(stream_name);
	g_free(client_name);

	self->updating = FALSE;
}

static void pama_source_output_widget_source_button_clicked(GtkButton *button, PamaSourceOutputWidget *self)
{
	GtkWidget *item, *last = NULL;
	GSList *sources, *iter;
	PamaPulseSource *source, *current_source;
	gchar *source_name;

	if (!self->source_menu)
	{
		self->source_menu = gtk_menu_new();
		g_object_ref_sink(self->source_menu);
		g_signal_connect(gtk_widget_get_toplevel(self->source_menu), "hide", G_CALLBACK(pama_source_output_widget_source_menu_hidden), self);

		sources = g_slist_copy(pama_pulse_context_get_sources(self->context));
		sources = g_slist_sort(sources, pama_pulse_source_compare_by_description);

		g_object_get(self->source_output, "source", &current_source, NULL);

		for (iter = sources; iter; iter = iter->next)
		{
			source = PAMA_PULSE_SOURCE(iter->data);
			g_object_get(source, "description", &source_name, NULL);

			if (last)
				item = gtk_radio_menu_item_new_with_label_from_widget(GTK_RADIO_MENU_ITEM(last), source_name);
			else
				item = gtk_radio_menu_item_new_with_label(NULL, source_name);

			g_object_set(item,
			             "active", (current_source == source),
			             "user-data", source,
			             NULL);
			last = item;
			g_free(source_name);

			g_signal_connect(item, "toggled", G_CALLBACK(pama_source_output_widget_source_menu_item_toggled), self);
			gtk_menu_shell_append(GTK_MENU_SHELL(self->source_menu), item);
			gtk_widget_show_all(item);
		}

		g_slist_free(sources);

		gtk_menu_popup(GTK_MENU(self->source_menu), NULL, NULL, NULL, NULL, 0, gtk_get_current_event_time());
	}
}
static void pama_source_output_widget_source_menu_hidden(GtkWidget *menu_window, PamaSourceOutputWidget *self)
{
	g_object_unref(self->source_menu);
	self->source_menu = NULL;
}
static void pama_source_output_widget_source_menu_item_toggled(GtkCheckMenuItem *item, PamaSourceOutputWidget *self)
{
	PamaPulseSource *source;

	if (!gtk_check_menu_item_get_active(item))
		return;

	g_object_get(item, "user-data", &source, NULL);

	pama_pulse_source_output_set_source(self->source_output, source);
}

gint pama_source_output_widget_compare(gconstpointer a, gconstpointer b)
{
	PamaSourceOutputWidget *A = PAMA_SOURCE_OUTPUT_WIDGET(a);
	PamaSourceOutputWidget *B = PAMA_SOURCE_OUTPUT_WIDGET(b);

	return pama_pulse_source_output_compare_by_name(A->source_output, B->source_output);
}
