/*
 * pama-source-widget.c: A widget to display and manipulate a source
 * Part of PulseAudio Mixer Applet
 * Copyright © Vassili Geronimos 2009 <v.geronimos@gmail.com>
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

 
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif
#include <glib.h>
#include <glib/gi18n.h>
 
#include "pama-source-widget.h"
#include "widget-settings.h"

static void     pama_source_widget_class_init(PamaSourceWidgetClass *klass);
static void     pama_source_widget_init(PamaSourceWidget *self);
static void     pama_source_widget_set_property(GObject *gobject, guint property_id, const GValue *value, GParamSpec *pspec);
static GObject* pama_source_widget_constructor(GType gtype, guint n_properties, GObjectConstructParam *properties);
static void     pama_source_widget_dispose(GObject *gobject);
static void     pama_source_widget_weak_ref_notify(gpointer data, GObject *where_the_object_was);

static void     pama_source_widget_source_notify   (GObject *gobject, GParamSpec *pspec, gpointer data);
static void     pama_source_widget_context_notify(GObject *gobject, GParamSpec *pspec, gpointer data);

static void     pama_source_widget_update_values  (PamaSourceWidget *self);
static void     pama_source_widget_default_toggled(GtkToggleButton *togglebutton, gpointer data);
static void     pama_source_widget_mute_toggled   (GtkToggleButton *togglebutton, gpointer data);
static void     pama_source_widget_volume_changed (GtkRange *range, gpointer data);

G_DEFINE_TYPE(PamaSourceWidget, pama_source_widget, GTK_TYPE_HBOX);


enum 
{
	PROP_0,

	PROP_SOURCE,
	PROP_CONTEXT,
	PROP_GROUP
};
enum
{
	REORDER_REQUEST_SIGNAL,
	LAST_SIGNAL
};
static guint widget_signals[LAST_SIGNAL] = {0,};

static void pama_source_widget_class_init(PamaSourceWidgetClass *klass)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
	GParamSpec *pspec;

	gobject_class->set_property = pama_source_widget_set_property;
	gobject_class->constructor  = pama_source_widget_constructor;
	gobject_class->dispose      = pama_source_widget_dispose;

	pspec = g_param_spec_object("source",
	                            "Pulse source",
	                            "The PamaPulseSource that is to be controlled.",
	                            PAMA_TYPE_PULSE_SOURCE,
	                            G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY | G_PARAM_STATIC_STRINGS);
	g_object_class_install_property(gobject_class, PROP_SOURCE, pspec);

	pspec = g_param_spec_object("context",
	                            "Pulse context",
	                            "The PamaPulseContext that is managing the current connection.",
	                            PAMA_TYPE_PULSE_CONTEXT,
	                            G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY | G_PARAM_STATIC_STRINGS);
	g_object_class_install_property(gobject_class, PROP_CONTEXT, pspec);

	pspec = g_param_spec_object("group",
	                            "Widget group",
	                            "Another source widget which this widget will be grouped with.",
	                            PAMA_TYPE_SOURCE_WIDGET,
	                            G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY | G_PARAM_STATIC_STRINGS);
	g_object_class_install_property(gobject_class, PROP_GROUP, pspec);

	widget_signals[REORDER_REQUEST_SIGNAL] =
		g_signal_new("reorder-request",
		             G_TYPE_FROM_CLASS(gobject_class),
		             G_SIGNAL_RUN_LAST,
		             0,
		             NULL,
		             NULL,
		             g_cclosure_marshal_VOID__VOID,
		             G_TYPE_NONE,
		             0);
}

static void pama_source_widget_init(PamaSourceWidget *self)
{
}

static void pama_source_widget_set_property(GObject *gobject, guint property_id, const GValue *value, GParamSpec *pspec)
{
	PamaSourceWidget *self = PAMA_SOURCE_WIDGET(gobject);

	switch(property_id)
	{
		case PROP_SOURCE:
			self->source = g_value_get_object(value);
			g_object_weak_ref(G_OBJECT(self->source), pama_source_widget_weak_ref_notify, self);
			break;

		case PROP_CONTEXT:
			self->context = g_value_get_object(value);
			g_object_weak_ref(G_OBJECT(self->context), pama_source_widget_weak_ref_notify, self);
			break;

		case PROP_GROUP:
			self->group = g_value_get_object(value);
			break;

		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(gobject, property_id, pspec);
			break;
	}
}
static GObject* pama_source_widget_constructor(GType gtype, guint n_properties, GObjectConstructParam *properties)
{
	GtkWidget *name, *alignment;
	GtkWidget *inner_box, *volume, *value, *mute, *default_source;

	GObject *gobject = G_OBJECT_CLASS(pama_source_widget_parent_class)->constructor(gtype, n_properties, properties);
	PamaSourceWidget *self = PAMA_SOURCE_WIDGET(gobject);

	if (NULL == self->source)
		g_error("An attempt was made to create a source widget with no source.");
	if (NULL == self->context)
		g_error("An attempt was made to create a source widget with no context.");

	name = g_object_new(GTK_TYPE_LABEL,
	                    "ellipsize", PANGO_ELLIPSIZE_MIDDLE,
	                    "wrap", FALSE,
	                    "width-chars", WIDGET_NAME_WIDTH_IN_CHARS,
	                    "xalign", 0.0f,
	                    NULL);
	gtk_box_pack_start(GTK_BOX(self), name, FALSE, FALSE, 0);
	self->name = name;

	alignment = gtk_alignment_new(0, 0.5, 0, 0);
	gtk_box_pack_start(GTK_BOX(self), alignment, FALSE, FALSE, 0);

	inner_box = gtk_hbox_new(FALSE, 6);
	gtk_container_add(GTK_CONTAINER(alignment), inner_box);

	volume = gtk_hscale_new_with_range(0, 100, 2.5); /* for scroll steps of 5% */
	gtk_scale_set_draw_value   (GTK_SCALE(volume), FALSE);
	gtk_widget_set_size_request(volume, WIDGET_VOLUME_SLIDER_WIDTH, -1);
	gtk_box_pack_start(GTK_BOX(inner_box), volume, FALSE, FALSE, 0);
	self->volume = volume;

	value = g_object_new(GTK_TYPE_LABEL,
	                     "width-chars", WIDGET_VALUE_WIDTH_IN_CHARS,
	                     "xalign", 1.0f,
	                     NULL);
	gtk_box_pack_start(GTK_BOX(inner_box), value, FALSE, FALSE, 0);
	self->value = value;

	mute = gtk_check_button_new();
	gtk_widget_set_tooltip_text(GTK_WIDGET(mute), _("Mute audio input from this device"));
	gtk_toggle_button_set_mode(GTK_TOGGLE_BUTTON(mute), FALSE);
	gtk_container_add(GTK_CONTAINER(mute), gtk_image_new_from_icon_name("audio-volume-muted", GTK_ICON_SIZE_MENU));
	gtk_box_pack_start(GTK_BOX(inner_box), mute, FALSE, FALSE, 0);
	self->mute = mute;

	if (self->group)
		default_source = gtk_radio_button_new_from_widget(GTK_RADIO_BUTTON(self->group->default_source));
	else
		default_source = gtk_radio_button_new(NULL);
	gtk_widget_set_tooltip_text(GTK_WIDGET(default_source), _("Set as default"));
	gtk_toggle_button_set_mode(GTK_TOGGLE_BUTTON(default_source), FALSE);
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(default_source), self->source == pama_pulse_context_get_default_source(self->context));
	gtk_container_add(GTK_CONTAINER(default_source), gtk_image_new_from_icon_name("emblem-default", GTK_ICON_SIZE_MENU));
	gtk_box_pack_start(GTK_BOX(inner_box), default_source, FALSE, FALSE, 0);
	self->default_source = default_source;

	pama_source_widget_update_values(self);

	g_signal_connect(volume,         "value-changed", G_CALLBACK(pama_source_widget_volume_changed),  self);
	g_signal_connect(mute,           "toggled",       G_CALLBACK(pama_source_widget_mute_toggled),    self);
	g_signal_connect(default_source, "toggled",       G_CALLBACK(pama_source_widget_default_toggled), self);

	self->source_notify_handler_id  = g_signal_connect(self->source,  "notify::volume",            G_CALLBACK(pama_source_widget_source_notify),     self);
	self->context_notify_handler_id = g_signal_connect(self->context, "notify::default-source-name", G_CALLBACK(pama_source_widget_context_notify),  self);

	self->group = NULL; /* don't need it for anything else */

	return gobject;
}
static void pama_source_widget_dispose(GObject *gobject)
{
	PamaSourceWidget *self = PAMA_SOURCE_WIDGET(gobject);

	if (self->source)
	{
		if (self->source_notify_handler_id)
		{
			g_signal_handler_disconnect(self->source, self->source_notify_handler_id);
			self->source_notify_handler_id = 0;
		}

		g_object_weak_unref(G_OBJECT(self->source), pama_source_widget_weak_ref_notify, self);
		self->source = NULL;
	}

	if (self->context)
	{
		if (self->context_notify_handler_id)
		{
			g_signal_handler_disconnect(self->context, self->context_notify_handler_id);
			self->context_notify_handler_id = 0;
		}

		g_object_weak_unref(G_OBJECT(self->context), pama_source_widget_weak_ref_notify, self);
		self->context = NULL;
	}

	G_OBJECT_CLASS(pama_source_widget_parent_class)->dispose(gobject);
}
static void pama_source_widget_weak_ref_notify(gpointer data, GObject *where_the_object_was)
{
	PamaSourceWidget *self = data;

	if ((GObject *)self->source    == where_the_object_was)
		self->source = NULL;

	if ((GObject *)self->context == where_the_object_was)
		self->context = NULL;

	/* widget is no longer usable without these items. */
	gtk_object_destroy(GTK_OBJECT(self));
}


static void pama_source_widget_source_notify(GObject *gobject, GParamSpec *pspec, gpointer data)
{
	PamaSourceWidget *self = data;
	pama_source_widget_update_values(self);
	g_signal_emit(self, widget_signals[REORDER_REQUEST_SIGNAL], 0);
}
static void pama_source_widget_context_notify(GObject *gobject, GParamSpec *pspec, gpointer data)
{
	PamaSourceWidget *self = data;

	self->updating = TRUE;
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(self->default_source), self->source == pama_pulse_context_get_default_source(self->context));
	self->updating = FALSE;
}

static void pama_source_widget_update_values(PamaSourceWidget *self)
{
	gchar    *temp;
	gchar    *description;
	guint     volume;
	gboolean  mute;

	self->updating = TRUE;

	g_object_get(self->source,
	             "description", &description,
	             "volume",      &volume,
	             "mute",        &mute,
	             NULL);

	temp = g_markup_printf_escaped("<b>%s</b>", description);
	gtk_label_set_markup(GTK_LABEL(self->name), temp);
	gtk_widget_set_tooltip_markup(GTK_WIDGET(self->name), temp);
	g_free(temp);
	g_free(description);

	temp = g_strdup_printf("%.0f%%", 100.0 * volume / PA_VOLUME_NORM);
	gtk_label_set_text(GTK_LABEL(self->value), temp);
	g_free(temp);

	gtk_range_set_value(GTK_RANGE(self->volume), 100.0 * volume / PA_VOLUME_NORM);
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(self->mute), mute);

	gtk_widget_set_sensitive(self->volume, !mute);
	gtk_widget_set_sensitive(self->value,  !mute);

	self->updating = FALSE;
}

static void pama_source_widget_default_toggled(GtkToggleButton *togglebutton, gpointer data)
{
	PamaSourceWidget *self = data;

	if (self->updating || !gtk_toggle_button_get_active(togglebutton))
		return;

	pama_pulse_source_set_as_default(self->source);
}
static void pama_source_widget_mute_toggled   (GtkToggleButton *togglebutton, gpointer data)
{
	PamaSourceWidget *self = data;

	if (self->updating)
		return;

	pama_pulse_source_set_mute(self->source, gtk_toggle_button_get_active(togglebutton));
}
static void pama_source_widget_volume_changed (GtkRange *range, gpointer data)
{
	PamaSourceWidget *self = data;

	if (self->updating)
		return;

	pama_pulse_source_set_volume(self->source, gtk_range_get_value(range) * PA_VOLUME_NORM / 100);
}


gint pama_source_widget_compare(gconstpointer a, gconstpointer b)
{
	PamaSourceWidget *A = PAMA_SOURCE_WIDGET(a);
	PamaSourceWidget *B = PAMA_SOURCE_WIDGET(b);

	return pama_pulse_source_compare_by_description(A->source, B->source);
}
